/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 *
 * You can obtain a copy of the license at /OPENSPML_V2_TOOLKIT.LICENSE
 * or http://www.openspml.org/v2/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at /OPENSPML_V2_TOOLKIT.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */
package org.openspml.v2.profiles;

import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.HashSet;
import java.util.Set;
import org.openspml.v2.profiles.xsd.XSDUnmarshaller;
import org.openspml.v2.util.Spml2Exception;
import org.openspml.v2.util.xml.ObjectFactory;

/**
 * This is the register class for the XSD profile. It only adds a an 
 * XSDUnmarchaller based in an XSD definition file. The entities can be
 * read from the XSD file or passed as an argument.
 * 
 * @author ricky
 */
public class XSDProfileRegistrar implements ObjectFactory.ProfileRegistrar {

    /**
     * Profile URI (urn:oasis:names:tc:SPML:2.0:profiles:XSD).
     */
    public static final String PROFILE_URI_STRING = "urn:oasis:names:tc:SPML:2.0:profiles:XSD";
    
    /**
     * The XSD URI String (http://www.w3.org/2001/XMLSchema).
     */
    public static String XSD_URI_STRING = "http://www.w3.org/2001/XMLSchema";
    
    /**
     * The name of this profile.
     */
    public static final String PROFILE_ID = "XPMLv2 XSD Profile";
    
    /**
     * The unmarshaller that is going to be registered and unregistered.
     */
    private XSDUnmarshaller xsdunm = null;
    
    /**
     * Constructor using the uRL of the XSD file and the entities managed
     * by this XSD profile.
     * @param url The URL for the XSD definition file.
     * @param entities The entities that this profile manages.
     * @throws Spml2Exception Some error parsing the XSD file
     */
    public XSDProfileRegistrar(URL url, Set<String> entities) throws Spml2Exception {
        xsdunm = new XSDUnmarshaller();
        xsdunm.initSchema(url);
        xsdunm.getEntities().addAll(entities);
    }
    
    /**
     * Constructor using just the XSD file. The entities are all the
     * elements defined in the XSD file.
     * @param url The XSD file.
     * @throws Spml2Exception Some error parsing the XSD file.
     */
    public XSDProfileRegistrar(URL url) throws Spml2Exception {
        this(url, new HashSet<String>());
        xsdunm.calculateElementInstances();
    }
    
    /**
     * Getter for the profile ID.
     * @return 
     */
    @Override
    public String getProfileId() {
        return PROFILE_ID;
    }

    /**
     * The profile URI String.
     * @return The profile URI
     * @throws Spml2Exception Some error
     */
    @Override
    public URI getProfileURI() throws Spml2Exception {
        try {
            return new URI(PROFILE_URI_STRING);
        }
        catch (URISyntaxException e) {
            throw new Spml2Exception(e);
        }
    }

    /**
     * Register the unmarshaller (no creator is used).
     * @param of The object factory to register to
     */
    @Override
    public void register(ObjectFactory of) {
        of.addOCEUnmarshaller(xsdunm);
    }

    /**
     * Unregister the unmarshaller.
     * @param of The object factory to unregister from
     */
    @Override
    public void unregister(ObjectFactory of) {
        of.removeOCEUnmarshaller(xsdunm);
    }
    
}
