/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 *
 * You can obtain a copy of the license at /OPENSPML_V2_TOOLKIT.LICENSE
 * or http://www.openspml.org/v2/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at /OPENSPML_V2_TOOLKIT.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */
package org.openspml.v2.profiles.xsd;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.net.URL;
import java.net.URLConnection;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.openspml.v2.msg.OpenContentElement;
import org.openspml.v2.util.Spml2Exception;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

/**
 * Class that represent the OpenCOntentElement as the DOM document.
 * It represents any XSD object that is sent or returned using the
 * XSD profile of the SPML standard.
 * 
 * @author ricky
 */
public class XSDDocElement implements OpenContentElement {

    /**
     * The document itself.
     */
    protected Document doc = null;
    
    /**
     * Empty constructor.
     */
    public XSDDocElement() {
        // noop
    }
    
    /**
     * Constructor using the doc itself.
     * @param doc The document that represents the object
     */
    public XSDDocElement(Document doc) {
        this.doc = doc;
    }
    
    /**
     * Assign the document using the XML representation in a string.
     * @param xsd The xsd representation
     * @throws Spml2Exception Some error
     */
    public XSDDocElement(String xsd) throws Spml2Exception {
        setXsdDoc(xsd);
    }
    
    /**
     * Assign the document using an URL (file or whatever) that contains the
     * XML text.
     * @param url The url of the XML document
     * @throws Spml2Exception Some error
     */
    public XSDDocElement(URL url) throws Spml2Exception {
        setXsdDoc(url);
    }

    /**
     * Return the DOM document representation of the object.
     * @return The document.
     */
    public Document getDocument() {
        return doc;
    }
    
    /**
     * Setter using a string XML.
     * @param xsd The xsd representation
     * @throws Spml2Exception Some error
     */
    public void setXsdDoc(String xsd) throws Spml2Exception {
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            this.doc = db.parse(new ByteArrayInputStream(xsd.getBytes()));
        } catch (ParserConfigurationException | SAXException | IOException e) {
            throw new Spml2Exception(e);
        }
    }
    
    /**
     * Setter using an external (file or whatever) XML source.
     * @param url The url of the XML document
     * @throws Spml2Exception Some error
     */
    public void setXsdDoc(URL url) throws Spml2Exception {
        try {
            URLConnection conn = url.openConnection();
            try (InputStream is = conn.getInputStream()) {
                DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                DocumentBuilder db = dbf.newDocumentBuilder();
                this.doc = db.parse(is);
            } catch (ParserConfigurationException | SAXException | IOException e) {
                throw new Spml2Exception(e);
            }
        } catch (IOException e) {
            throw new Spml2Exception(e);
        }
    }
    
    /**
     * The toXML implementation just uses the transformer from DOM to XML.
     * @param indent Not used.
     * @return The XML representation of the Document
     * @throws Spml2Exception Some error
     */
    @Override
    public String toXML(int indent) throws Spml2Exception {
        try {
            TransformerFactory tf = TransformerFactory.newInstance();
            Transformer transformer = tf.newTransformer();
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
            StringWriter writer = new StringWriter();
            transformer.transform(new DOMSource(doc), new StreamResult(writer));
            String xsd = writer.getBuffer().toString();
            return xsd;
        } catch (TransformerException e) {
            throw new Spml2Exception(e);
        }
    }

    /**
     * Same method than the previous one.
     * @return The XML representation of the Document
     * @throws Spml2Exception Some error
     */
    @Override
    public String toXML() throws Spml2Exception {
        return toXML(0);
    }
}
