/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 *
 * You can obtain a copy of the license at /OPENSPML_V2_TOOLKIT.LICENSE
 * or http://www.openspml.org/v2/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at /OPENSPML_V2_TOOLKIT.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */
package org.openspml.v2.profiles.xsd;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;
import org.openspml.v2.msg.OpenContentElement;
import org.openspml.v2.profiles.XSDProfileRegistrar;
import org.openspml.v2.util.Spml2Exception;
import org.openspml.v2.util.xml.ObjectFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * Unmarshaller that from the DOM object creates a XSDDocElement for all
 * the XSD objects.
 * 
 * @author ricky
 */
public class XSDUnmarshaller implements ObjectFactory.OCEUnmarshaller {

    /**
     * The name of the entities managed by the XSD defined by the profile.
     */
    private Set<String> entities = null;
    
    /**
     * The XSD schema validator in DOM to check read objects.
     */
    private Schema schema = null;
    
    /**
     * The XSD schema as an element.
     */
    XSDDocElement xsd = null;
    
    
    private class ListErrorHandler implements ErrorHandler {
        
        List<Exception> error = new ArrayList<>();
        List<Exception> warning = new ArrayList<>();

        public boolean hasErrors() {
            return !error.isEmpty();
        }
        
        public Exception[] getErrors() {
            return error.toArray(new Exception[0]);
        }
        
        public boolean hasWarnings() {
            return !warning.isEmpty();
        }
        
        public Exception[] getWarnings() {
            return warning.toArray(new Exception[0]);
        }
        
        @Override
        public void warning(SAXParseException saxpe) throws SAXException {
            warning.add(saxpe);
        }

        @Override
        public void error(SAXParseException saxpe) throws SAXException {
            error.add(saxpe);
        }

        @Override
        public void fatalError(SAXParseException saxpe) throws SAXException {
            error.add(saxpe);
        }
        
    }
    
    public XSDUnmarshaller() {
        entities = new HashSet();
        schema = null;
    }
    
    /**
     * Initializes the schema using a file
     * @param file 
     * @throws org.openspml.v2.util.Spml2Exception Error creating the schema
     */
    public void initSchema(File file) throws Spml2Exception {
        try {
            String language = XMLConstants.W3C_XML_SCHEMA_NS_URI;
            SchemaFactory factory = SchemaFactory.newInstance(language);
            schema = factory.newSchema(file);
            xsd = new XSDDocElement(file.toURI().toURL());
        } catch (SAXException|MalformedURLException e) {
            throw new Spml2Exception(e);
        }
    }
    
    /**
     * Initializes the schema using a file 
     * @param url The url for the xsd
     * @throws org.openspml.v2.util.Spml2Exception Error creating the schema
     */
    public void initSchema(URL url) throws Spml2Exception {
        try {
            String language = XMLConstants.W3C_XML_SCHEMA_NS_URI;
            SchemaFactory factory = SchemaFactory.newInstance(language);
            schema = factory.newSchema(url);
            xsd = new XSDDocElement(url);
        } catch (SAXException e) {
            throw new Spml2Exception(e);
        }
    }
    
    /**
     * Access to the managed entities of the XSD profile.
     * @return 
     */
    public Set<String> getEntities() {
        return this.entities;
    }
    
    public String getTargetNamespace() {
        return xsd.getDocument().getDocumentElement().getAttribute("targetNamespace");
    }
    
    public void calculateElementInstances() {
        NodeList list = xsd.getDocument().getDocumentElement().getChildNodes();
        for (int i = 0; i < list.getLength(); i++) {
            Node node = list.item(i);
            if (node.getNodeType() == Node.ELEMENT_NODE && ("element".equals(node.getNodeName()) || "element".equals(node.getLocalName()))) {
                Node attr = list.item(i).getAttributes().getNamedItem("name");
                if (attr != null) {
                    this.entities.add(attr.getNodeValue());
                }
            }
        }
    }
    
    /**
     * Main method that marshalls  the schema and the different XSD objects
     * that can be sent or received. The idea is using the namespace of the
     * XSD to know if they are XSD objects.
     * 
     * @param obj The DOM object to unmarshall
     * @return The XSDDocElement associated to the XSD object
     * @throws Spml2Exception Some error
     */
    @Override
    public OpenContentElement unmarshall(Object obj) throws Spml2Exception {
        System.err.println("XSDUnmarshaller.unmarshall");
        System.err.println("obj: " + obj);
        if (obj instanceof Node) {
            Node node = (Node) obj;
            System.err.println("Namespace: " + node.getNamespaceURI());
            if (node.getNodeType() == Node.ELEMENT_NODE
                    && XSDProfileRegistrar.XSD_URI_STRING.equals(node.getNamespaceURI())) {
                // unmarshall the XSD schema itself
                try {
                    DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                    DocumentBuilder builder = dbf.newDocumentBuilder();
                    Document doc = builder.newDocument();
                    Node newNode = node.cloneNode(true);
                    doc.adoptNode(newNode);
                    doc.appendChild(newNode);
                    return new XSDDocElement(doc);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            } else if (node.getNodeType() == Node.ELEMENT_NODE
                    && entities.contains(node.getLocalName()) && getTargetNamespace().equals(node.getNamespaceURI())) {
                // unmarshall a XML object (for example add request)
                try {
                    DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                    DocumentBuilder builder = dbf.newDocumentBuilder();
                    Document doc = builder.newDocument();
                    Node newNode = node.cloneNode(true);
                    doc.adoptNode(newNode);
                    doc.appendChild(newNode);
                    Validator validator = schema.newValidator();
                    validator.setErrorHandler(null);
                    ListErrorHandler eh = new ListErrorHandler();
                    validator.setErrorHandler(eh);
                    validator.validate(new DOMSource(doc));
                    if (!eh.hasErrors()) {
                        return new XSDDocElement(doc);
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                }
            } else if (node.getNodeType() == Node.ELEMENT_NODE && getTargetNamespace().equals(node.getNamespaceURI())) {
                // unmarchall the XML fragment (for example modify request)
                // for modification I'm requesting the URI of the XSD
                // which is not specified in the standard
                try {
                    DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                    DocumentBuilder builder = dbf.newDocumentBuilder();
                    Document doc = builder.newDocument();
                    Node newNode = node.cloneNode(true);
                    doc.adoptNode(newNode);
                    doc.appendChild(newNode);
                    return new XSDDocElement(doc);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }
        return null;
    }

    
}
